//===========================================================================
//	CMaxMidiSMF Class Implementation
//
//	(C) Copyright, Paul A. Messick, 1996
//===========================================================================
#include "stdafx.h"
#include "MaxMidi.h"

//---------------------------------------------------------------------------
//	CMaxMidiSMF Constructors
//---------------------------------------------------------------------------
CMaxMidiSMF::CMaxMidiSMF()
{
	fIsOpen = FALSE;
	nTracksInSMF = 0;
	nTracksAttached = 0;
	pTrackList = NULL;
}

CMaxMidiSMF::CMaxMidiSMF(LPSTR filename, const char Mode)
{
	fIsOpen = FALSE;
	nTracksInSMF = 0;
	nTracksAttached = 0;
	pTrackList = NULL;

	Open(filename, Mode);
}

CMaxMidiSMF::CMaxMidiSMF(LPCTSTR filename, const char Mode)
{
	CMaxMidiSMF((LPSTR)filename, Mode);
}

//---------------------------------------------------------------------------
//	Attach/Detach Track
//---------------------------------------------------------------------------
void CMaxMidiSMF::Attach(CMaxMidiTrack* pTrack)
{
	// attach this track to the end of the list
	Attach(pTrack, -1);
}

void CMaxMidiSMF::Attach(CMaxMidiTrack* pTrack, int position)
{
	int i;

	// adjust the format type if more than one track is attached
	nTracksAttached++;
	if(nTracksAttached > 1)
		Format = 1;

	// expand the track list for one more entry
	if(pTrackList)
		// add an entry if list exists
		pTrackList = (CMaxMidiTrack**)GlobalReAlloc(pTrackList, nTracksAttached * sizeof(CMaxMidiTrack*), GPTR | GMEM_MOVEABLE);
	else
		// create the list
		pTrackList = (CMaxMidiTrack**)GlobalAlloc(GPTR, sizeof(CMaxMidiTrack*));
		
	// append this track to the end of the track list if the position is -1
	if(position == -1)
		pTrackList[nTracksAttached - 1] = pTrack;
	else
	// move the other tracks up by one and insert this track at position
	{
		for(i = nTracksAttached - 1; i >= position; i--)
			pTrackList[i + 1] = pTrackList[i];

		pTrackList[position] = pTrack;
	}

	pTrack->Attach(this);
}

BOOL CMaxMidiSMF::Detach(CMaxMidiTrack* pTrack)
{
	int i, j;

	// make sure there is at least one track to detach
	if(nTracksAttached == 0)
		return FALSE;

	// search for the specified track in the track list
	for(i = 0; i < nTracksAttached; i++)
	{
		if(pTrackList[i] == pTrack)
		{
			// move all of the following tracks down by one entry
			// thus removing the offending track
			nTracksAttached--;
			for(j = i; j < nTracksAttached; j++)
				pTrackList[j] = pTrackList[j + 1];

			// shrink the list
			if(nTracksAttached == 0)
			{
				GlobalFree(pTrackList);
				pTrackList = NULL;
			}
			else
				pTrackList = (CMaxMidiTrack**)GlobalReAlloc(pTrackList, nTracksAttached * sizeof(CMaxMidiTrack*), GPTR | GMEM_MOVEABLE);

			// it's removed!
			return TRUE;
		}
	}

	// else, this track is not in the list!
	return FALSE;
}

//---------------------------------------------------------------------------
//	CMaxMidiSMF Destructor
//---------------------------------------------------------------------------
CMaxMidiSMF::~CMaxMidiSMF()
{
	Close();
}

//---------------------------------------------------------------------------
//	Open
//---------------------------------------------------------------------------
BOOL CMaxMidiSMF::Open(LPCTSTR filename, const char Mode, int Format)
{
	return Open((LPSTR)filename, Mode, Format);
}

BOOL CMaxMidiSMF::Open(LPSTR filename, const char Mode, int Format)
{
	// make sure there are no open files attached to this object
	Close();

	// open the file
	CMaxMidiSMF::Mode = Mode;
	CMaxMidiSMF::Format = Format;
	if((hSMF = OpenSMF(filename, &(CMaxMidiSMF::Format), Mode, &nTracksInSMF)) != 0)
		fIsOpen = TRUE;

	return fIsOpen;
}

//---------------------------------------------------------------------------
//	Close
//---------------------------------------------------------------------------
void CMaxMidiSMF::Close(void)
{
	if(fIsOpen)
	{
		CloseSMF(hSMF);

		// detach any tracks from this file
		while(nTracksAttached)
		{
			pTrackList[0]->Detach(this);
			Detach(pTrackList[0]);
		}

		nTracksInSMF = 0;
	}

	fIsOpen = FALSE;
}

//---------------------------------------------------------------------------
//	SMF Functions
//---------------------------------------------------------------------------
BOOL CMaxMidiSMF::Read(CMaxMidiTrack* pTrack)
{
	DWORD dwEventsRead = 0;
	DWORD dwRead;
	int Track;
	DWORD dwTrackSize;
	LPMIDIEVENT pBuffer;

	// only valid if opened for read
	if(Mode == READ && pTrackList)
	{
		// find the track number corresponding to this track
		for(Track = 0; Track < nTracksAttached; Track++)
			if(pTrackList[Track] == pTrack)
			{
				dwTrackSize = pTrack->GetBufferSize();
				pBuffer = pTrack->GetBuffer();

				do {
					// expand the track buffer, if necessary
					if(dwTrackSize - dwEventsRead < DEFAULT_BUFFER_SIZE)
					{
						dwTrackSize += dwEventsRead;
						pBuffer = (LPMIDIEVENT)GlobalReAlloc(pBuffer, dwTrackSize * sizeof(MidiEvent), GPTR | GMEM_MOVEABLE);
					}

					// read a block of events
					dwRead = ReadSMF(hSMF, Track, &pBuffer[dwEventsRead], DEFAULT_BUFFER_SIZE);
					dwEventsRead += dwRead;

				} while(dwRead);

				pTrack->SetNumEvents(dwEventsRead);
				pTrack->SetBufferSize(dwTrackSize);
				pTrack->SetBuffer(pBuffer);
				pTrack->GetName();
				break;
			}
	}

	return dwEventsRead != 0;
}

BOOL CMaxMidiSMF::Write(CMaxMidiTrack* pTrack)
{
	int Track;
	LPSTR lpName;

	// only valid if opened for write
	if(Mode == WRITE && pTrackList)
	{
		// find the track number corresponding to this track
		for(Track = 0; Track < nTracksAttached; Track++)
			if(pTrackList[Track] == pTrack)
			{
				if(lpName = pTrack->GetName())
					WriteMeta(pTrack, META_NAME, lpName, 0);

				return (WriteSMF(hSMF, Track, pTrack->GetBuffer(), pTrack->GetNumEvents()) == pTrack->GetNumEvents());
			}
	}

	return FALSE;
}

BOOL CMaxMidiSMF::Save(void)
{
	int i;
	CMaxMidiIn* pMidiIn;
	CMaxMidiSync* pSync;

	// only valid if opened for write
	if(Mode != WRITE || pTrackList == NULL)
		return FALSE;

	// get the resolution sync device, if attached, and set it in the file
	for(i = 0; i < nTracksAttached; i++)
		if(pMidiIn = pTrackList[i]->GetMidiIn())
			if(pSync = pMidiIn->GetSync())
			{
				CMaxMidiSMF::Resolution(pSync->Resolution());
				break;
			}

	// save all of the tracks that are attached
	for(i = 0; i < nTracksAttached; i++)
		Write(pTrackList[i]);

	return TRUE;
}

BOOL CMaxMidiSMF::Load(void)
{
	int i;
	CMaxMidiSync* pSync;
	CMaxMidiOut* pMidiOut;

	// only valid if opened for read
	if(Mode != READ || pTrackList == NULL)
		return FALSE;

	// read all of the tracks that are attached
	RewindSMF(hSMF);
	for(i = 0; i < nTracksAttached; i++)
	{
		pTrackList[i]->Flush();
		Read(pTrackList[i]);
	}

	// get the resolution from the file and set it in the sync device, if attached
	if(pMidiOut = pTrackList[0]->GetMidiOut())
		if(pSync = pMidiOut->GetSync())
			pSync->Resolution(CMaxMidiSMF::Resolution());

	return TRUE;
}

//---------------------------------------------------------------------------
//	Meta Event Functions
//---------------------------------------------------------------------------
DWORD CMaxMidiSMF::ReadMeta(CMaxMidiTrack* pTrack, BYTE type, LPSTR* Value, DWORD* cbSize)
{
	int Track;
	LPSTR valbuf;
	DWORD valbufsize;
	DWORD time = (DWORD)-1;

	// *** V1.56 DLL ***
	*cbSize = 0;

	// only valid if opened for read
	if(Mode == READ && pTrackList)
	{
		// find the track number corresponding to this track
		for(Track = 0; Track < nTracksAttached; Track++)
			if(pTrackList[Track] == pTrack)
			{
				// look for the event in the track
				time = ReadMetaEvent(hSMF, Track, type, &valbuf, &valbufsize);
				if(time != (DWORD)-1)
				{
					// create a buffer for the event value and copy it over
					*Value = (LPSTR)GlobalAlloc(GPTR, valbufsize);
					memcpy(*Value, valbuf, valbufsize);

					// *** V1.56 DLL ***
					*cbSize = valbufsize;
				}
			}
	}

	return time;
}

BOOL CMaxMidiSMF::WriteMeta(CMaxMidiTrack* pTrack, BYTE type, LPSTR Value, DWORD time)
{
	int Track;
	UINT rc = 1;

	// only valid if opened for write
	if(Mode == WRITE && pTrackList)
	{
		// find the track number corresponding to this track
		for(Track = 0; Track < nTracksAttached; Track++)
			if(pTrackList[Track] == pTrack)
				rc = WriteMetaEvent(hSMF, Track, type, Value, time);
	}

	return rc == 0;
}
