//===========================================================================
//	CMaxMidiTrack Class Implementation
//
//	(C) Copyright, Paul A. Messick, 1996
//===========================================================================
#include "stdafx.h"
#include "MaxMidi.h"

//---------------------------------------------------------------------------
//	CMaxMidiTrack Constructors
//---------------------------------------------------------------------------
CMaxMidiTrack::CMaxMidiTrack()
{
	lpBuffer = NULL;
	lpName = NULL;
#ifdef _SMF
	pSMF = NULL;
#endif
#ifdef _MIDIOUT
	pMidiOut = NULL;
#endif
#ifdef _MIDIIN
	pMidiIn = NULL;
#endif
	inPtr = outPtr = 0;
	IsRecording(FALSE);
	Mute(FALSE);

	CreateBuffer();
}

//---------------------------------------------------------------------------
//	CMaxMidiTrack Destructor
//---------------------------------------------------------------------------
CMaxMidiTrack::~CMaxMidiTrack()
{
	// detach the track from any smf, input, or output devices
	Detach();
	FreeBuffer();
}

//---------------------------------------------------------------------------
//	Attach/Detach
//---------------------------------------------------------------------------
#ifdef _SMF
void CMaxMidiTrack::Attach(CMaxMidiSMF* pSMF)
{
	CMaxMidiTrack::pSMF = pSMF;
}

void CMaxMidiTrack::Detach(CMaxMidiSMF* pSMF)
{
	if(pSMF == CMaxMidiTrack::pSMF)
		CMaxMidiTrack::pSMF = NULL;
}
#endif

#ifdef _MIDIOUT
void CMaxMidiTrack::Attach(CMaxMidiOut* pMidiOut)
{
	CMaxMidiTrack::pMidiOut = pMidiOut;
}

void CMaxMidiTrack::Detach(CMaxMidiOut* pMidiOut)
{
	if(pMidiOut == CMaxMidiTrack::pMidiOut)
		pMidiOut = NULL;
}
#endif

#ifdef _MIDIIN
void CMaxMidiTrack::Attach(CMaxMidiIn* pMidiIn)
{
	CMaxMidiTrack::pMidiIn = pMidiIn;
}

void CMaxMidiTrack::Detach(CMaxMidiIn* pMidiIn)
{
	if(pMidiIn == CMaxMidiTrack::pMidiIn)
		pMidiIn = NULL;
}
#endif

void CMaxMidiTrack::Detach(void)
{
	// detach from any attached devices
#ifdef _SMF
	if(pSMF)
	{
		pSMF->Detach(this);
		pSMF = NULL;
	}
#endif
#ifdef _MIDIOUT
	if(pMidiOut)
	{
		pMidiOut->Detach(this);
		pMidiOut = NULL;
	}
#endif
#ifdef _MIDIIN
	if(pMidiIn)
	{
		pMidiIn->Detach(this);
		pMidiIn = NULL;
		IsRecording(FALSE);
	}
#endif
}

//---------------------------------------------------------------------------
//	Track Name
//---------------------------------------------------------------------------
LPSTR CMaxMidiTrack::GetName(void)
{
	if(lpName)
		return lpName;

#ifdef _SMF
	DWORD cbSize;

	// get the name from the smf, if it exists
	if(pSMF && pSMF->GetMode() == READ)
	{
		pSMF->ReadMeta(this, META_NAME, &lpName, &cbSize);
		return lpName;
	}
#endif

	// currently no name
	return NULL;
}

void CMaxMidiTrack::SetName(LPSTR name)
{
	// free the old name, if there was one
	if(lpName)
		GlobalFree(lpName);

	// allocate space for the new name
	// *** V1.50 ***
	// need to allocate extra byte for the trailing null
	lpName = (LPSTR)GlobalAlloc(GPTR, strlen(name) + 1);
	strcpy(lpName, name);
}

//---------------------------------------------------------------------------
//	Buffer Functions
//---------------------------------------------------------------------------
LPMIDIEVENT CMaxMidiTrack::GetEvent(DWORD eventNum)
{
	if(eventNum >= dwBufSize)
		return NULL;

	return &lpBuffer[eventNum];
}

void CMaxMidiTrack::SetEvent(LPMIDIEVENT lpEvent, DWORD eventNum)
{
	if(eventNum >= dwBufSize)
		return;

	lpBuffer[eventNum] = *lpEvent;
}

BOOL CMaxMidiTrack::CreateBuffer(DWORD dwBufEvents)
{
	// this can only be called once per open
	if(lpBuffer)
		return FALSE;

	// create buffers for data from each track
	lpBuffer = (LPMIDIEVENT)GlobalAlloc(GPTR, sizeof(MidiEvent) * dwBufEvents);
	dwBufSize = dwBufEvents;

	return TRUE;
}

void CMaxMidiTrack::FreeBuffer(void)
{
	if(lpBuffer)
		GlobalFree(lpBuffer);

	lpBuffer = NULL;
}

//---------------------------------------------------------------------------
//	Queued Read/Write
//---------------------------------------------------------------------------
LPMIDIEVENT CMaxMidiTrack::Read(void)
{
	// don't do anything if at end of buffer, or recording or muted
	if(outPtr == inPtr || fMute || fRecord)
		return NULL;

	return &lpBuffer[outPtr++];
}

void CMaxMidiTrack::Write(LPMIDIEVENT lpEvent)
{
	// don't write into the track if not recording
	if(!fRecord)
		return;

	InsertEvent(lpEvent, (DWORD)-1);
}

//---------------------------------------------------------------------------
//	Event Editing Functions
//---------------------------------------------------------------------------
void CMaxMidiTrack::InsertEvent(LPMIDIEVENT lpEvent, DWORD beforeEvent)
{
	// does the buffer need to grow in order to add this event?
	if(inPtr == dwBufSize)
	{
		dwBufSize += BUFFER_GROW_SIZE;
		lpBuffer = (LPMIDIEVENT)GlobalReAlloc(lpBuffer, dwBufSize * sizeof(MidiEvent), GPTR | GMEM_MOVEABLE);
	}

	// insert at the end (append) if beforeEvent is -1
	if(beforeEvent == (DWORD)-1)
	{
		// put it in the buffer
		lpBuffer[inPtr] = *lpEvent;
		inPtr++;
		return;
	}

	// else, move all of the events following this new one up by one slot
	memmove(&lpBuffer[beforeEvent + 1], &lpBuffer[beforeEvent], inPtr - beforeEvent);

	// add the event to the track by putting in the specified location
	// without adjusting the timestamp of the following event
	lpBuffer[beforeEvent] = *lpEvent;
	inPtr++;	
}

void CMaxMidiTrack::DeleteEvent(DWORD eventNum)
{
	// eventNum == -1 will delete the last event
	inPtr--;
	if(eventNum < inPtr)
	{
		// move all of the events following the specfied event down
		// in the track by one, thus deleteing the event
		memmove(&lpBuffer[eventNum], &lpBuffer[eventNum + 1], inPtr - eventNum);
	}
}

void CMaxMidiTrack::SlideTrack(DWORD eventNum, int delta)
{
	// add the delta time to the specified event
	if(eventNum < inPtr)
	{
		lpBuffer[eventNum].time += delta;
		if((int)lpBuffer[eventNum].time < 0)
			lpBuffer[eventNum].time = 0;
	}
}

//---------------------------------------------------------------------------
//	Save/Load
//---------------------------------------------------------------------------
BOOL CMaxMidiTrack::Load(void)
{
#ifdef _SMF
	// if the SMF object is attached, read data into this track
	if(pSMF)
		return pSMF->Read(this);
#endif

	return FALSE;
}

BOOL CMaxMidiTrack::Save(void)
{
#ifdef _SMF
	// if the SMF object is attached, save this track
	// this assumes that the buffer is properly attached to the smf object
	// NOTE: this should only be done in ascending order of tracks
	// muted tracks are not written
	if(pSMF && !fMute)
	{
		// write the name as the first event in the track, if it exists
		if(lpName)
			pSMF->WriteMeta(this, META_NAME, lpName, 0);

		return pSMF->Write(this);
	}
#endif

	// if not attached, return false
	return FALSE;
}

//---------------------------------------------------------------------------
//	Absolute Time/Delta Time Functions
//---------------------------------------------------------------------------
LPMIDIEVENT CMaxMidiTrack::GetAbsBuffer(DWORD startEvent, DWORD* numEvents)
{
	LPMIDIEVENT lpBuf;

	// sanity check: make sure the requested start event is in the buffer
	// and the track is not muted or recording
	if(startEvent >= inPtr || fMute || fRecord)
	{
		*numEvents = 0;
		return NULL;
	}

	// the buffer size will be either the requested number of events
	// or the number remaining in the track, whichever is lower
	*numEvents = ((inPtr - startEvent) < *numEvents ? (inPtr - startEvent) : *numEvents);

	// create the buffer
	lpBuf = (LPMIDIEVENT)GlobalAlloc(GPTR, *numEvents * sizeof(MidiEvent));
	memcpy(lpBuf, &lpBuffer[startEvent], *numEvents * sizeof(MidiEvent));

	// convert it to absolute time
	DeltaToAbs(lpBuf, startEvent, *numEvents);

	return lpBuf;
}

DWORD CMaxMidiTrack::AbsNow(DWORD eventNum)
{
	DWORD time;
	int i;

	// sanity check: make sure the specified event is in the buffer
	// if not, return zero as the time
	if(eventNum >= inPtr)
		return 0;

	// add up the elapsed number of ticks since the beginning
	// of the track till this event
	time = lpBuffer[0].time;
	for(i = 1; i <= (int)eventNum; i++)
		time += lpBuffer[i].time;

	return time;
}

void CMaxMidiTrack::AbsToDelta(LPMIDIEVENT lpBuf, DWORD startEvent, DWORD numEvents)
{
	DWORD now;
	DWORD start;
	int i;

	// get the abs time of the previous event (since we need the delta from it
	// for the time for the first event is this buffer)
	// if this is the first event in the track (i.e. startEvent == 0) then we know
	// the start time is zero
	if(startEvent == 0)
		start = 0;
	else
		start = AbsNow(startEvent - 1);

	// convert each abs time to delta time by subtracting the abs time from
	// the start time of the event prior event
	for(i = 0; i < (int)numEvents; i++)
	{
		// save the abs time of this event
		// it will become the start for the next event
		now = lpBuffer[i].time;
		lpBuffer[i].time -= start;
		start = now;
	}
}

void CMaxMidiTrack::DeltaToAbs(LPMIDIEVENT lpBuf, DWORD startEvent, DWORD numEvents)
{
	DWORD now;
	int i;

	// get the abs time of the first event in the buffer
	now = AbsNow(startEvent);

	// we know the abs time for the first one -- its the startEvent
	lpBuf[0].time = now;

	// convert each time stamp to abs by setting it to the 
	// time since the beginning of the track
	for(i = 1; i < (int)numEvents; i++)
	{
		now += lpBuf[i].time;
		lpBuf[i].time = now;
	}
}

DWORD CMaxMidiTrack::GetTime(DWORD eventNum)
{
	// sanity check: make sure the specified event is in the buffer
	// if not, return zero as the time
	if(eventNum >= inPtr)
		return 0;

	// return the time for the requested event in track
	return lpBuffer[eventNum].time;
}

